#include <cfloat>
#include "camera.h"

void Camera::
render_image(Array2<Vec3f>& image,
             float supersampling) const
{
   // first invert the model-matrix transformation to speed up transforming
   // rays from camera space to world space.
   Transformation mvinv=inverse(modelview);

   // Loop over all samples, tracing rays as we go.
   Array1<Vec2f> pixel_xy;
   make_uniform_samples(supersampling, pixel_xy);
   //make_jittered_samples(supersampling, pixel_xy); // this is optional!
   Array1<Vec3f> sample_color(pixel_xy.size());
   for(int k=0; k<(int)pixel_xy.size(); ++k){
      float pixel_x=pixel_xy[k][0];
      float pixel_y=pixel_xy[k][1];
      Ray ray=generate_ray(mvinv, pixel_x, pixel_y);
      // here's where we actually do raytracing for this sample
      sample_color[k]=raytracer.trace_ray(ray);
      // provide some feedback on the progress of the raytracer
      // (technically not a great idea in a parallel loop, but adequate for now)
      if(k%20000==19999) std::cout<<" "<<k*100./pixel_xy.size()<<"% done"<<std::endl;
      else if(k%500==499) std::cout<<"."<<std::flush;
   }
   std::cout<<std::endl;

   // reconstruct image from samples
   std::cout<<"reconstructing image from samples..."<<std::endl;
   // black out the image, resizing it to the desired dimensions
   image.assign(pixel_width, pixel_height, Vec3f(0,0,0));
   Array2f weight(pixel_width, pixel_height, 0.f);
   for(unsigned int k=0; k<pixel_xy.size(); ++k){
      float pixel_x=pixel_xy[k][0];
      float pixel_y=pixel_xy[k][1];
      int pi=(int)std::floor(pixel_x);
      int pj=(int)std::floor(pixel_y);
      float subi=pixel_x-pi;
      float subj=pixel_y-pj;
      float u0, u1, u2, v0, v1, v2;
      quadratic_bspline_weights(subi, u0, u1, u2);
      quadratic_bspline_weights(subj, v0, v1, v2);
      if(pi>0){
         if(pj>0){
            weight(pi-1,pj-1)+=u0*v0;
            image(pi-1,pj-1)+=(u0*v0)*sample_color[k];
         }
         weight(pi-1,pj)+=u0*v1;
         image(pi-1,pj)+=(u0*v1)*sample_color[k];
         if(pj+1<pixel_height){
            weight(pi-1,pj+1)+=u0*v2;
            image(pi-1,pj+1)+=(u0*v2)*sample_color[k];
         }
      }
      if(pj>0){
         weight(pi,pj-1)+=u1*v0;
         image(pi,pj-1)+=(u1*v0)*sample_color[k];
      }
      weight(pi,pj)+=u1*v1;
      image(pi,pj)+=(u1*v1)*sample_color[k];
      if(pj+1<pixel_height){
         weight(pi,pj+1)+=u1*v2;
         image(pi,pj+1)+=(u1*v2)*sample_color[k];
      }
      if(pi+1<pixel_width){
         if(pj>0){
            weight(pi+1,pj-1)+=u2*v0;
            image(pi+1,pj-1)+=(u2*v0)*sample_color[k];
         }
         weight(pi+1,pj)+=u2*v1;
         image(pi+1,pj)+=(u2*v1)*sample_color[k];
         if(pj+1<pixel_height){
            weight(pi+1,pj+1)+=u2*v2;
            image(pi+1,pj+1)+=(u2*v2)*sample_color[k];
         }
      }
   }
   std::cout<<"normalizing..."<<std::flush;
   for(int j=0; j<pixel_height; ++j) for(int i=0; i<pixel_width; ++i)
      if(weight(i,j)) image(i,j)/=weight(i,j);
   std::cout<<" done."<<std::endl;
}

Ray Camera::
generate_ray(const Transformation& modelview_inverse,
      float pixel_x,
             float pixel_y) const
{
   Vec3f cam_origin=
      near*Vec3f(left+(pixel_x/pixel_width)*(right-left),
                 bottom+(pixel_y/pixel_height)*(top-bottom),
                 -1);
   Vec3f cam_direction(cam_origin);
   Ray ray(modelview_inverse.apply_to_point(cam_origin),
           modelview_inverse.apply_to_direction(cam_direction));
   normalize(ray.direction);
   return ray;
}

void Camera::
make_uniform_samples(float supersampling,
                     Array1<Vec2f>& pixel_xy) const
{
   const int si=(int)(pixel_width*supersampling);
   const int sj=(int)(pixel_height*supersampling);
   pixel_xy.resize(si*sj);
   int k=0;
   for(int j=0; j<sj; ++j){
      for(int i=0; i<si; ++i){
         pixel_xy[k][0]=pixel_width*(i+0.5f)/si;
         pixel_xy[k][1]=pixel_height*(j+0.5f)/sj;
         ++k;
      }
   }
}

void Camera::
make_jittered_samples(float supersampling,
                      Array1<Vec2f>& pixel_xy) const
{
   unsigned int seed=0;
   const int si=(int)(pixel_width*supersampling);
   const int sj=(int)(pixel_height*supersampling);
   pixel_xy.resize(si*sj);
   int k=0;
   for(int j=0; j<sj; ++j){
      for(int i=0; i<si; ++i){
         // create a jittered sample
         pixel_xy[k][0]=pixel_width*(i+randhashf(seed++))/si;
         pixel_xy[k][1]=pixel_height*(j+randhashf(seed++))/sj;
         ++k;
      }
   }
}
