#include <cstdarg>
#include <cstdlib>
#include <cmath>
#include <fstream>
#include "mesh_io.h"

#define LINESIZE 4096 // maximum line size when reading .obj files

bool
write_objfile(const std::vector<Vec3i>& tri,
              const std::vector<Vec3f>& x,
              const std::vector<Vec3f>* normal,
              const char* filename_format,
              ...)
{
#ifdef WIN32
   va_list ap;
   va_start(ap, filename_format);
   int len=_vscprintf(filename_format, ap) // _vscprintf doesn't count
                                       +1; // terminating '\0'
   char* filename=new char[len];
   vsprintf(filename, filename_format, ap);
   std::ofstream output(filename, std::ofstream::out);
   delete[] filename;
   va_end(ap);
#else
   va_list ap;
   va_start(ap, filename_format);
   char* filename;
   vasprintf(&filename, filename_format, ap);
   std::ofstream output(filename, std::ofstream::out);
   std::free(filename);
   va_end(ap);
#endif

   if(!output.good()) return false;

   for(unsigned int i=0; i<x.size(); ++i)
      output<<"v "<<x[i]<<std::endl;
   if(normal && !normal->empty()){
      for(unsigned int i=0; i<x.size(); ++i)
         output<<"vn "<<(*normal)[i]<<std::endl;
   }
   for(unsigned int t=0; t<tri.size(); ++t)
      output<<"f "<<tri[t][0]+1<<' '<<tri[t][1]+1<<' '<<tri[t][2]+1<<std::endl; // correct for 1-based indexing in OBJ files
   return output.good();
}

static bool
read_int(const char* s,
         int& value,
         bool& leading_slash,
         int& position)
{
   leading_slash=false;
   for(position=0; s[position]!=0; ++position){
      switch(s[position]){
         case '/':
            leading_slash=true;
            break;
         case '0': case '1': case '2': case '3': case '4':
         case '5': case '6': case '7': case '8': case '9':
            goto found_int;
      }
   }
   return false;
   
   found_int:
   value=0;
   for(;; ++position){
      switch(s[position]){
         case '0': case '1': case '2': case '3': case '4':
         case '5': case '6': case '7': case '8': case '9':
            value=10*value+s[position]-'0';
            break;
         default:
            return true;
      }
   }
   return true; // should never get here, but keeps compiler happy
}

static void
read_face_list(const char* s,
               std::vector<int>& vertex_list)
{
   vertex_list.clear();
   int v, skip;
   bool leading_slash;
   for(int i=0;;){
      if(read_int(s+i, v, leading_slash, skip)){
         if(!leading_slash)
            vertex_list.push_back(v-1); // correct for 1-based index
         i+=skip;
      }else
         break;
   }
}

bool
read_objfile(std::vector<Vec3i>& tri,
             std::vector<Vec3f>& x,
             std::vector<Vec3f>* normal,
             const char *filename_format,
             ...)
{
#ifdef WIN32
   va_list ap;
   va_start(ap, filename_format);
   int len=_vscprintf(filename_format, ap) // _vscprintf doesn't count
                                       +1; // terminating '\0'
   char* filename=new char[len];
   vsprintf(filename, filename_format, ap);
   std::ifstream input(filename, std::ifstream::in);
   delete[] filename;
   va_end(ap);
#else
   va_list ap;
   va_start(ap, filename_format);
   char* filename;
   vasprintf(&filename, filename_format, ap);
   std::ifstream input(filename, std::ifstream::in);
   std::free(filename);
   va_end(ap);
#endif

   if(!input.good()) return false;

   tri.clear();
   x.clear();
   if(normal) normal->clear();

   char line[LINESIZE];
   std::vector<int> vertex_list;
   while(input.good()){
      input.getline(line, LINESIZE);
      switch(line[0]){
         case 'v': // vertex data
            if(line[1]==' '){
               Vec3f new_vertex;
               std::sscanf(line+2, "%f %f %f",
                           &new_vertex[0], &new_vertex[1], &new_vertex[2]);
               x.push_back(new_vertex);
            }else if(line[1]=='n' && line[2]==' '){
               if(normal){
                  Vec3f new_normal;
                  std::sscanf(line+2, "%f %f %f",
                              &new_normal[0], &new_normal[1], &new_normal[2]);
                  normal->push_back(new_normal);
               }
            }
            break;
         case 'f': // face data
            if(line[1]==' '){
               read_face_list(line+2, vertex_list);
               // automatically (and blindly) triangulate
               // - doing this to minimize edge lengths would be a bit better
               for(int j=0; j<(int)vertex_list.size()-2; ++j)
                  tri.push_back(Vec3i(vertex_list[0],
                                      vertex_list[j+1],
                                      vertex_list[j+2]));
            }
            break;
      }
   }
   return true;
}
