#ifndef PRIMITIVES_H
#define PRIMITIVES_H

// Basic types and operations for the raytracer: rays and bounding boxes.

#include "vec.h"

// A ray consists of two vectors, one being the origin point and the other a
// direction. It's not enforced in this class, but it's strongly recommended
// for everyone's sanity to stick with unit-length normalized directions.
struct Ray
{
   Vec3f origin, direction;

   Ray(void) : origin(0,0,0), direction(0,0,-1) {}

   Ray(const Vec3f& origin_,
       const Vec3f& direction_)
      : origin(origin_), direction(direction_)
   {}
};

// an IO operator so you can more easily dump out a ray when debugging
inline std::ostream&
operator<<(std::ostream &out,
           const Ray& ray)
{
   out<<"(o:"<<ray.origin<<", d:"<<ray.direction<<")";
   return out;
}

// An axis-aligned bounding box, represented with two vectors---the minimum
// coordinate corner and the maximum coordinate corner.
struct BoundingBox
{
   Vec3f xmin, xmax;

   // the default is a zero-extent box at the origin
   BoundingBox(void) : xmin(0,0,0), xmax(0,0,0) {}

   // or if you give it a single point, a zero-extent box at the point
   BoundingBox(const Vec3f& x) : xmin(x), xmax(x) {}

   // if you construct it with two vectors, it assumes they are the min
   // and the max corners
   BoundingBox(const Vec3f& xmin_,
               const Vec3f& xmax_)
      : xmin(xmin_), xmax(xmax_)
   { assert(xmin[0]<=xmax[0] && xmin[1]<=xmax[1] && xmin[2]<=xmax[2]); }

   // you can also build a box from a list of three points, finding the
   // min and max for each axis.
   void
   build_from_points(const Vec3f& x0,
                     const Vec3f& x1,
                     const Vec3f& x2)
   { minmax(x0, x1, x2, xmin, xmax); }

   bool
   contains(const Vec3f& x) const
   {
      return xmin[0]<=x[0] && x[0]<=xmax[0]
          && xmin[1]<=x[1] && x[1]<=xmax[1]
          && xmin[2]<=x[2] && x[2]<=xmax[2];
   }

   // you can enlarge an existing box to include a new point (if it doesn't
   // contain it already)
   void
   include_point(const Vec3f& x)
   { update_minmax(x, xmin, xmax); }

   // or you can even enlarge it to include another box (if it doesn't contain
   // it already)
   void
   include_box(const BoundingBox& box)
   {
      for(unsigned int i=0; i<3; ++i){
         if(box.xmin[i]<xmin[i]) xmin[i]=box.xmin[i];
         if(box.xmax[i]>xmax[i]) xmax[i]=box.xmax[i];
      }
   }

   // increases the bounding box by a factor of 1+epsilon in each axis:
   // useful for dealing with rounding errors!
   void
   enlarge(float epsilon)
   {
      float a=(epsilon/2)*(xmax[0]-xmin[0]);
      xmin[0]-=a; xmax[0]+=a;
      float b=(epsilon/2)*(xmax[1]-xmin[1]);
      xmin[1]-=b; xmax[1]+=b;
      float c=(epsilon/2)*(xmax[2]-xmin[2]);
      xmin[2]-=c; xmax[2]+=c;
   }

   // determine if the ray intersects the bounding box within [smin,smax]
   bool
   intersect(const Ray& ray,
             float smin,
             float smax) const
   { return intersect_and_update(ray, smin, smax); }

   // return if the ray intersects the bounding box within the range
   // [smin,smax], clipping them to the range of the ray intersecting the box
   bool
   intersect_and_update(const Ray& ray,
                        float& smin,
                        float& smax) const
   {
      // loop over axes, shrinking (tmin,tmax) by intersecting ray with
      // each slab
      for(int axis=0; axis<3; ++axis){
         if(ray.direction[axis]==0){ // is ray parallel to this slab?
            // if the ray doesn't start in the slab it's parallel to,
            // it can't intersect the box
            if(ray.origin[axis]<xmin[axis] || ray.origin[axis]>xmax[axis])
               return false;
         }else{ // if not, we can look for intersections
            float s0=-(ray.origin[axis]-xmin[axis])/ray.direction[axis],
                  s1=-(ray.origin[axis]-xmax[axis])/ray.direction[axis];
            if(s1<s0) std::swap(s0,s1);
            if(s0>smin) smin=s0;
            if(s1<smax) smax=s1;
            if(smin>smax) return false; // if interval empty, no intersection
         }
      }
      return true;
   }

   Vec3f
   world_space(float fraction0,
               float fraction1,
               float fraction2)
   {
      return Vec3f((1-fraction0)*xmin[0]+fraction0*xmax[0],
                   (1-fraction1)*xmin[1]+fraction1*xmax[1],
                   (1-fraction2)*xmin[2]+fraction2*xmax[2]);
   }
};

inline BoundingBox
box_intersection(const BoundingBox& a,
                 const BoundingBox& b)
{
   return BoundingBox(max_union(a.xmin, b.xmin), min_union(a.xmax, b.xmax));
}

// an IO operator so you can more easily dump out a box when debugging
inline std::ostream&
operator<<(std::ostream &out,
           const BoundingBox& box)
{
   out<<"("<<box.xmin<<" ~ "<<box.xmax<<")";
   return out;
}

#endif
