#include "shape.h"

void MultiShape::
build_tree(void)
{
   std::vector<BoundingBox> box(subshape.size());
   for(unsigned int i=0; i<subshape.size(); ++i)
      box[i]=subshape[i]->bounding_box();
   tree.construct_from_leaf_boxes(box.size(), &box[0]);
}

bool MultiShape::
any_intersection(const Ray& ray,
                 float smin,
                 float smax) const
{
   // first check if the ray hits the root's bounding box
   if(!tree.box.intersect_and_update(ray, smin, smax))
      return false;
   Array1<const BoundingBoxTree*> stack;
   stack.push_back(&tree);
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i)
         if(subshape[node->index[i]]->any_intersection(ray, smin, smax))
            return true;
      // check for children nodes (with boxes intersecting the ray)
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, smin, smax))
            stack.push_back(node->children[i]); 
      }
   }
   return false;
}

bool MultiShape::
first_intersection(const Ray& ray,
                   float smin,
                   float smax,
                   float& s,
                   Vec3f& normal) const
{
   // first check if the ray hits the root's bounding box
   if(!tree.box.intersect_and_update(ray, smin, smax))
      return false;
   bool found_intersection=false;
   Array1<const BoundingBoxTree*> stack; // ideally use a heap instead
   stack.push_back(&tree);
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         if(subshape[node->index[i]]->first_intersection(ray, smin, smax,
                                                         s, normal)){
            found_intersection=true;
            smax=s; // eliminate any more distant intersections
         }
      }
      // children nodes to check?
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, smin, smax))
            stack.push_back(node->children[i]); 
      }
   }
   return found_intersection;
}

bool MultiShape::
first_intersection(const Ray& ray,
                   float smin,
                   float smax,
                   float& s) const
{
   // first check if the ray hits the root's bounding box
   if(!tree.box.intersect_and_update(ray, smin, smax))
      return false;
   bool found_intersection=false;
   Array1<const BoundingBoxTree*> stack; // ideally use a heap instead
   stack.push_back(&tree);
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         if(subshape[node->index[i]]->first_intersection(ray, smin, smax, s)){
            found_intersection=true;
            smax=s; // eliminate any more distant intersections
         }
      }
      // children nodes to check?
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, smin, smax))
            stack.push_back(node->children[i]); 
      }
   }
   return found_intersection;
}

bool MultiShape::
contains(const Vec3f& x) const
{
   // first check if the ray hits the root's bounding box
   if(!tree.box.contains(x))
      return false;
   Array1<const BoundingBoxTree*> stack; // ideally use a heap instead
   stack.push_back(&tree);
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i)
         if(subshape[node->index[i]]->contains(x)) return true;
      // children nodes to check?
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.contains(x))
            stack.push_back(node->children[i]); 
      }
   }
   return false;
}
