#include "triangle_mesh.h"

void TriangleMesh::
build_tree(void)
{
   std::vector<BoundingBox> box(tri.size());
   for(unsigned int f=0; f<tri.size(); ++f){
      int i, j, k; assign(tri[f], i, j, k);
      box[f].build_from_points(x[i], x[j], x[k]);
      // enlarge to avoid rounding problems later
      float maxlength=max(dist(x[i],x[j]), dist(x[i],x[k]), dist(x[j],x[k]));
      float epsilon=1e-6*maxlength;
      box[f].xmin-=Vec3f(epsilon);
      box[f].xmax+=Vec3f(epsilon);
   }
   tree.construct_from_leaf_boxes(tri.size(), &box[0]);
}

bool TriangleMesh::
any_intersection(const Ray& ray,
                 float smin,
                 float smax) const
{
   // first check if the ray hits the root's bounding box
   if(!tree.box.intersect_and_update(ray, smin, smax))
      return false;
   // if we get here it does hit; do a depth first search of the tree
   std::vector<const BoundingBoxTree*> stack;
   stack.push_back(&tree);
   float s, alpha, beta, gamma;
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any primitive geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         if(intersect_triangle(node->index[i], ray, smin, smax, s,
                               alpha, beta, gamma)){
            return true;
         }
      }
      // check if there are children nodes (with boxes intersecting the ray)
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, smin, smax))
            stack.push_back(node->children[i]); 
      }
   }
   return false;
}

bool TriangleMesh::
first_intersection(const Ray& ray,
                   float smin,
                   float smax,
                   float& s,
                   Vec3f& normal) const
{
   // first check if the ray hits the root's bounding box
   if(!tree.box.intersect_and_update(ray, smin, smax)) return false;
   // if we get here it hits tree; depth first search for first intersection
   float s0 = s;
   bool found_intersection=false;
   std::vector<const BoundingBoxTree*> stack;
   stack.push_back(&tree);
   float alpha, beta, gamma;
   while(!stack.empty()){
      const BoundingBoxTree *node=stack.back();
      stack.pop_back();
      // check if there's any primitive geometry to check in this node
      for(unsigned int i=0; i<node->index.size(); ++i){
         int f=node->index[i];
         if(intersect_triangle(f, ray, smin, smax, s0, alpha, beta, gamma)){
            smax=s0; // eliminate any further intersections from consideration
            normal=normalized(cross(x[tri[f][1]]-x[tri[f][0]],
                                    x[tri[f][2]]-x[tri[f][0]]));
            found_intersection=true;
         }
      }
      // children nodes (with boxes intersecting the ray) to search later?
      for(unsigned int i=0; i<node->children.size(); ++i){
         if(node->children[i]->box.intersect(ray, smin, smax))
            stack.push_back(node->children[i]); 
      }
   }
   s=s0;
   return found_intersection;
}

BoundingBox TriangleMesh::
bounding_box(void) const
{
   assert(!tree.index.empty() || !tree.children.empty());
   return tree.box;
}

// This is the core issue: how to intersect a ray with a triangle.
bool TriangleMesh::
intersect_triangle(unsigned int f,
                   const Ray& ray,
                   float smin,
                   float smax,
                   float& s,
                   float& alpha,
                   float& beta,
                   float& gamma) const
{
   // get the vertex indices (i, j, k) from tri[f]: the triangle has vertices
   // x[i], x[j], and x[k]
   int i, j, k; assign(tri[f], i, j, k);

   // Here is where you need to fill in code to test if ray intersects this
   // triangle in the parameter range smin<s<smax
   // If it does, set the parameter value s and the barycentric coordinates
   // alpha, beta and gamma of the intersection, and return true.
   // If it doesn't, return false but DO NOT MODIFY s, alpha, beta or gamma.
   // @@@@@@@@@@@@@
   return false; // delete this line once you have your version filled in
}
